/*
 * Copyright (C) 2009 Ericsson AB
 *
 * Authors: Torgny Johansson <torgny.johansson@ericsson.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>
#include <glib.h>
#include <glib/gthread.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib.h>

#include <fcntl.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <locale.h>
#include <glib/gi18n.h>

#include "config.h"
#include "gps_control.h"
#include "gps_control_log_window.h"
#include "gps_control_supl_certificate.h"

#include "gps_gui_settings.h"

#define NMEA_DEV "/dev/gps2"
#define MAX_BUF_SIZE 32
#define MAX_POS_SIZE 20
#define MAX_NMEA_SIZE 512
#define MAX_MSG_SIZE 1024

#define ICON_DISCONNECTED "/usr/share/mbm-gps-control/GPSDisconnected.png"
#define ICON_CONNECTED  "/usr/share/mbm-gps-control/GPSConnected.png"

#define GPS_NO_FIX_TIMEOUT 300

#define SUPL_NI_ALLOW_ON_TIMEOUT 0
#define SUPL_NI_DENY_ON_TIMEOUT 1
#define SUPL_NI_NOTIFICATION 2

#define MBM_DBUS_SERVICE "org.mbm.MbmGps"
#define MBM_DBUS_PATH "/org/mbm/Gps"

struct MBMGuiPrivate{
	GtkWidget *window;
	GtkWidget *notebook;
	GtkWidget *settings_vbox;
	GtkWidget *start_button;
	GtkWidget *stop_button;
	GtkWidget *single_shot_checkbutton;
	GtkWidget *mode_combo_box;
	GtkWidget *interval_spin_button;
	GtkWidget *assisted_alignment;
	GtkWidget *supl_ni_dialog;
	GtkWidget *supl_uncert_dialog;
	GtkWidget *systray_menu;
	GtkStatusIcon *status_icon;
	GtkWidget *statusbar;
	GtkWidget *status_image;
	guint statusbar_context;
	
	GtkWidget *settings_frame;
	GtkWidget *supl_settings_vbox;
	GtkWidget *enable_supl_radio_button;
	GtkWidget *supl_apn_entry;
	GtkWidget *supl_server_entry;
	GtkWidget *supl_server_auto_checkbutton;
	GtkWidget *supl_user_entry;
	GtkWidget *supl_password_entry;
	GtkWidget *supl_allow_ni_checkbutton;
	
	GtkWidget *latitude_label;
	GtkWidget *longitude_label;
	GtkWidget *altitude_label;
	GtkWidget *fix_quality_label;
	
	fd_set fdsread;
	fd_set fdsreaduse;
	int nmea_fd;
	int maxfd;
	struct timeval timeout;

	int valid_fix;
	
	DBUS_Connection g_conn;
	int connection_failed;
	int current_uncert_msg_id;
	int current_ni_request_msg_id;
	int supl_ni_dialog_showing;
	int supl_uncert_dialog_showing;
	int accept_unknown_certificate;
	int old_gps_mode;
	int supl_failed;
        gboolean supl_selected;
	
	gboolean no_fix_dialog_displayed;
};

static struct MBMGuiPrivate *priv;

static void show_supl_uncert_dialog (char *message, int message_id);
static void show_supl_ni_dialog (char *message, int mode, int message_id);
static void show_connection_failed_dialog ();
static void setup_settings_frame (void);

extern void view_log_window (GtkButton * button, gpointer user_data)
{
	show_log_window ();
}

extern void view_install_cert_window (GtkButton * button, gpointer user_data)
{
	supl_cert_show_window ();
}

static void show_connection_failed_dialog ()
{
	GtkWidget *dialog;
	char *message =
		_
		("Connection to the mbm-gpsd daemon failed.\nMake sure mbm-gpsd is running then restart the client.");
	priv->connection_failed = 1;
	dialog =
		gtk_message_dialog_new (GTK_WINDOW (priv->window), GTK_DIALOG_MODAL,
								GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
								"%s", message);
	gtk_window_set_title (GTK_WINDOW (dialog),
						  _("Connection to daemon failed"));
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_main_quit ();
	exit (1);
}

static int send_supl_ni_reply (int fd, int msg_id, int allow)
{
	char buf[MAX_BUF_SIZE];
	int written;
	
	sprintf (buf, "AT*E2GPSSUPLNIREPLY=%d,%d\r\n", msg_id, allow);
	written = write (fd, buf, strlen (buf));
	return written;
}

static int send_supl_uncert_reply (int fd, int msg_id, int allow)
{
	char buf[MAX_BUF_SIZE];
	int written;
	
	sprintf (buf, "AT*E2CERTUNREPLY=%d,%d\r\n", msg_id, allow);
	written = write (fd, buf, strlen (buf));
	return written;
}

static void open_nmea_fd ()
{
	int status;
	struct termios settings;

	priv->nmea_fd = open (NMEA_DEV, O_RDWR);
	if (priv->nmea_fd == -1) {
		g_error ("Open of %s failed: %m", NMEA_DEV);
		exit (1);
	}

	FD_SET (priv->nmea_fd, &priv->fdsread);
	priv->fdsreaduse = priv->fdsread;
	if (priv->nmea_fd >= priv->maxfd)
		priv->maxfd = priv->nmea_fd + 1;

	tcflush (priv->nmea_fd, TCIOFLUSH);
	status = tcgetattr (priv->nmea_fd, &settings);
	if (status) {
		g_error ("getting settings failed: %m");
		exit (1);
	}
	settings.c_lflag = ISIG | ICANON | IEXTEN | ECHOE | ECHOK;
	cfsetispeed (&settings, B4800);
	cfsetospeed (&settings, B4800);
	status = tcsetattr (priv->nmea_fd, TCSANOW, &settings);
	if (status) {
		g_error ("setting baudrate failed: %m");
		exit (1);
	}
}

static void parse_nmea_data (char *nmea_data, int len)
{
	gchar **nmea_infos;
	gchar latitude[MAX_POS_SIZE];
	gchar longitude[MAX_POS_SIZE];
	gchar altitude[MAX_POS_SIZE];
	char *str;
	int b, c;

	nmea_infos = g_strsplit (nmea_data, ",", 15);
	if (strstr (nmea_infos[0], "$GPGGA")) {
		priv->valid_fix = atoi (nmea_infos[6]);
		if (priv->valid_fix) {
			gtk_label_set_text (GTK_LABEL (priv->fix_quality_label), _("Valid"));
			g_stpcpy (latitude, nmea_infos[2]);
			g_strlcat (latitude, _(" N"), sizeof (latitude));
			gtk_label_set_text (GTK_LABEL (priv->latitude_label), latitude);

			g_stpcpy (longitude, nmea_infos[4]);
			g_strlcat (longitude, _(" E"), sizeof (longitude));
			gtk_label_set_text (GTK_LABEL (priv->longitude_label), longitude);

			g_stpcpy (altitude, nmea_infos[9]);
			g_strlcat (altitude, _(" M"), sizeof (altitude));
			gtk_label_set_text (GTK_LABEL (priv->altitude_label), altitude);
		} else {
			gtk_label_set_text (GTK_LABEL (priv->fix_quality_label), _("Invalid"));
			gtk_label_set_text (GTK_LABEL (priv->latitude_label), _("N/A"));
			gtk_label_set_text (GTK_LABEL (priv->longitude_label), _("N/A"));
			gtk_label_set_text (GTK_LABEL (priv->altitude_label), _("N/A"));
		}
		return;
	}

	str = strstr (nmea_infos[0], "*E2CERTUN:");
	if (str) {
		b = atoi (nmea_infos[1]);
		c = atoi (nmea_infos[2]);
		if (c == 1) {
			if (!priv->accept_unknown_certificate) {
				priv->current_uncert_msg_id = b;
				show_supl_uncert_dialog (_
										 ("SUPL server not certified.\n\nContinue with secure connection?"),
										 b);
			} else {
				priv->current_uncert_msg_id = b;
				send_supl_uncert_reply (priv->nmea_fd, b, 1);
			}
		}
		return;
	}

	str = strstr (nmea_infos[0], "*E2GPSSUPLNI:");
	if (str) {
		b = atoi (nmea_infos[1]);	/* msg id */
		c = atoi (nmea_infos[2]);	/* msg type */

		if (c == SUPL_NI_NOTIFICATION) {
			/* show notification that location has been given out */
			priv->current_ni_request_msg_id = -1;
			show_supl_ni_dialog (_
								 ("A location request has been received from the network.\nYour location has been given out."),
								 c, b);
		} else if (c == SUPL_NI_DENY_ON_TIMEOUT
				   || c == SUPL_NI_ALLOW_ON_TIMEOUT) {
			/* show verification dialog, deny/allow on timeout */
			priv->current_ni_request_msg_id = b;
			show_supl_ni_dialog (_
								 ("A location request has been received from the network.\nDo you want to give out your position?"),
								 c, b);
		}
		return;
	}
}

static gboolean read_nmea (gpointer user_data)
{
	int select_res;
	char buf[MAX_NMEA_SIZE];
	int len;

	if (priv->connection_failed)
		return FALSE;

	/* read nmea data */
	memset (buf, 0, MAX_NMEA_SIZE);
	priv->timeout.tv_sec = 0;
	priv->timeout.tv_usec = 100;
	priv->fdsreaduse = priv->fdsread;
	select_res = select (priv->maxfd, &priv->fdsreaduse, NULL, NULL, &priv->timeout);
	if (select_res != 0) {
		len = read (priv->nmea_fd, buf, sizeof (buf));

		if (strstr (buf, "$GP") && (len > 0)) {
			log_window_append (buf, len);
			parse_nmea_data (buf, len);
		}

		if (strstr (buf, "*") && (len > 0))
			parse_nmea_data (buf, len);
	}

	return TRUE;
}

void dbus_connect (DBUS_Connection * conn)
{
	GError *error;

#ifndef TARGET_ALTLINUX
	g_type_init ();
#endif

	error = NULL;
	conn->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
	if (conn->connection == NULL) {
		g_printerr ("Failed to open connection to bus: %s\n", error->message);
		g_error_free (error);
		exit (1);
	}

	/* Create a proxy object for the "bus driver"  */

	conn->proxy = dbus_g_proxy_new_for_name (conn->connection, MBM_DBUS_SERVICE,
											 MBM_DBUS_PATH, MBM_DBUS_SERVICE);
}

gchar *dbus_getter_str (DBUS_Connection * conn, gchar * method)
{
	char *strret;
	GError *error = NULL;
	if (!dbus_g_proxy_call (conn->proxy, method, &error, G_TYPE_INVALID,
							G_TYPE_STRING, &strret, G_TYPE_INVALID))
		show_connection_failed_dialog ();

	return strret;
	/* remember to do g_free (strret); */
}

void dbus_setter_str (DBUS_Connection * conn, gchar * method, char *value)
{
	GError *error = NULL;

	if (!dbus_g_proxy_call (conn->proxy, method, &error,
							G_TYPE_STRING, value, G_TYPE_INVALID,
							G_TYPE_INVALID))
		show_connection_failed_dialog ();
}

gint dbus_getter_int (DBUS_Connection * conn, gchar * method)
{
	gint intret;
	GError *error = NULL;

	if (!dbus_g_proxy_call (conn->proxy, method, &error, G_TYPE_INVALID,
							G_TYPE_INT, &intret, G_TYPE_INVALID))
		show_connection_failed_dialog ();
	
	return intret;
}

void dbus_setter_int (DBUS_Connection * conn, gchar * method, gint value)
{
	GError *error = NULL;

	if (!dbus_g_proxy_call (conn->proxy, method, &error,
							G_TYPE_INT, value, G_TYPE_INVALID, G_TYPE_INVALID))
		show_connection_failed_dialog ();
}

void dbus_disconnect (DBUS_Connection * conn)
{
	g_object_unref (conn->proxy);
}

static gboolean check_time_without_fix (gpointer user_data)
{
	gint seconds_without_fix;
	GtkWidget *dialog;
	gint response;
	DBUS_Connection *conn = (DBUS_Connection *) user_data;

	if (priv->no_fix_dialog_displayed)
		return TRUE;
	
	seconds_without_fix = dbus_getter_int (conn, "GetTimeWithoutFix");

	if ((seconds_without_fix >= GPS_NO_FIX_TIMEOUT)) {
		dialog = gtk_message_dialog_new (GTK_WINDOW (priv->window),
										 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
										 GTK_MESSAGE_WARNING,
										 GTK_BUTTONS_NONE,
										 "%s",
										 _("Unable to obtain GPS fix.\n\nPlease make sure the system has a clear view of the sky."));

		gtk_dialog_add_buttons (GTK_DIALOG (dialog),
								_("Continue"),
								GTK_RESPONSE_ACCEPT,
								_("Stop GPS"),
								GTK_RESPONSE_REJECT,
								NULL);

		gtk_window_set_title (GTK_WINDOW (dialog), _("No fix obtained"));

		response = gtk_dialog_run (GTK_DIALOG (dialog));

		switch (response)
		{
		case GTK_RESPONSE_REJECT:
			/* stop the gps */
			stop_gps ();
			break;
		default:
			/* do nothing */
			break;
		}
		gtk_widget_destroy (dialog);
		priv->no_fix_dialog_displayed = TRUE;
	}

	return TRUE;
}

static gboolean update_gps_status (gpointer user_data)
{
	gboolean enabled = FALSE;
	gint gps_mode = 0;
	gint pref_gps_mode = 0;
	DBUS_Connection *conn = (DBUS_Connection *) user_data;
	char sa_message[128];
	
	strcpy (sa_message, _("GPS is on in standalone mode "));

	if (priv->connection_failed)
		return FALSE;

	enabled = dbus_getter_int (conn, "GetGpsEnabled");
	gps_mode = dbus_getter_int (conn, "GetEnabledGpsMode");
	pref_gps_mode = dbus_getter_int (conn, "GetPrefGpsMode");

	if ((gps_mode == GPS_MODE_S_ALONE) && (priv->old_gps_mode == GPS_MODE_SUPL)
		&& (pref_gps_mode == GPS_MODE_SUPL))
		priv->supl_failed = 1;

	if ((gps_mode == GPS_MODE_S_ALONE) && (pref_gps_mode == GPS_MODE_S_ALONE))
		priv->supl_failed = 0;

	if (gps_mode == GPS_MODE_S_ALONE_SUPL_FAILED)
		priv->supl_failed = 1;

	if ((gps_mode == GPS_MODE_S_ALONE) && priv->supl_selected)
		priv->supl_failed = 1;

	gtk_statusbar_pop (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context);
	if (enabled == GPS_MODE_WAKING_UP) {
		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
							_("The GPS is waking up. Please wait."));
		gtk_widget_set_sensitive (priv->start_button, FALSE);
		gtk_widget_set_sensitive (priv->stop_button, FALSE);
		gtk_image_set_from_file (GTK_IMAGE (priv->status_image), ICON_DISCONNECTED);
		gtk_status_icon_set_from_file (priv->status_icon, ICON_DISCONNECTED);
	} else if (enabled == GPS_MODE_RECOVERING) {
		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
							_
							("The GPS has stalled. Please wait while recovering."));
		gtk_widget_set_sensitive (priv->start_button, FALSE);
		gtk_widget_set_sensitive (priv->stop_button, FALSE);
		gtk_image_set_from_file (GTK_IMAGE (priv->status_image), ICON_DISCONNECTED);
		gtk_status_icon_set_from_file (priv->status_icon, ICON_DISCONNECTED);
	} else if (enabled == GPS_MODE_SUPL_WAITING_FOR_REGISTRATION) {
		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
							_
							("Waiting for network registration for SUPL."));
		gtk_widget_set_sensitive (priv->start_button, FALSE);
		gtk_widget_set_sensitive (priv->stop_button, TRUE);
		gtk_image_set_from_file (GTK_IMAGE (priv->status_image), ICON_DISCONNECTED);
		gtk_status_icon_set_from_file (priv->status_icon, ICON_DISCONNECTED);
	} else if (enabled) {
		if (gps_mode == GPS_MODE_SUPL) {
			gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
								_("GPS is on in SUPL mode"));
		} else {
			if (priv->supl_failed)
				strcat (sa_message, _("(SUPL failed)"));
			
			gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
								sa_message);
		}
		gtk_widget_set_sensitive (priv->start_button, FALSE);
		gtk_widget_set_sensitive (priv->stop_button, TRUE);
		gtk_widget_set_sensitive (priv->single_shot_checkbutton, FALSE);
		gtk_image_set_from_file (GTK_IMAGE (priv->status_image), ICON_CONNECTED);
		gtk_status_icon_set_from_file (priv->status_icon, ICON_CONNECTED);
	} else {
		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
							_("GPS is off"));
		gtk_label_set_text (GTK_LABEL (priv->fix_quality_label), _("N/A"));
		gtk_widget_set_sensitive (priv->start_button, TRUE);
		gtk_widget_set_sensitive (priv->stop_button, FALSE);
		gtk_widget_set_sensitive (priv->single_shot_checkbutton, TRUE);
		gtk_image_set_from_file (GTK_IMAGE (priv->status_image), ICON_DISCONNECTED);
		gtk_status_icon_set_from_file (priv->status_icon, ICON_DISCONNECTED);
	}

	priv->old_gps_mode = gps_mode;
	return TRUE;
}

static void set_systray_menu_position (GtkMenu * menu, gint * x, gint * y,
									   gboolean * push_in, gpointer user_data)
{
	gtk_status_icon_position_menu (GTK_MENU (priv->systray_menu),
								   x,
								   y,
								   push_in,
								   priv->status_icon);
}

static void set_supl_settings_sensitivity (gboolean sensitivity)
{
	gtk_widget_set_sensitive (priv->supl_apn_entry, sensitivity);
	gtk_widget_set_sensitive (priv->supl_server_auto_checkbutton, sensitivity);
	
	if (sensitivity) {
		if (!gtk_toggle_button_get_active
			(GTK_TOGGLE_BUTTON (priv->supl_server_auto_checkbutton)))
			gtk_widget_set_sensitive (priv->supl_server_entry, sensitivity);
		else
			gtk_widget_set_sensitive (priv->supl_server_entry, FALSE);
	} else
		gtk_widget_set_sensitive (priv->supl_server_entry, FALSE);

	gtk_widget_set_sensitive (priv->supl_user_entry, sensitivity);
	gtk_widget_set_sensitive (priv->supl_password_entry, sensitivity);
	gtk_widget_set_sensitive (priv->supl_allow_ni_checkbutton, sensitivity);
}

extern void on_supl_apn_entry_activate (GtkEntry * entry, gpointer user_data)
{
	dbus_setter_str (&priv->g_conn, "SetSuplApn", (char *)gtk_entry_get_text (entry));
}

extern void on_supl_server_entry_activate (GtkEntry * entry, gpointer user_data)
{
	dbus_setter_str (&priv->g_conn, "SetSuplAddress",
					 (char *)gtk_entry_get_text (entry));
}

extern void on_supl_user_entry_activate (GtkEntry * entry, gpointer user_data)
{
	dbus_setter_str (&priv->g_conn, "SetSuplUser",
					 (char *)gtk_entry_get_text (entry));
}

extern void on_supl_password_entry_activate (GtkEntry * entry,
											 gpointer user_data)
{
	dbus_setter_str (&priv->g_conn, "SetSuplPassword",
					 (char *)gtk_entry_get_text (entry));
}

extern void on_supl_allow_ni_checkbutton_toggled (GtkToggleButton *
												  togglebutton,
												  gpointer user_data)
{
	if (gtk_toggle_button_get_active (togglebutton))
		dbus_setter_int (&priv->g_conn, "SetSuplEnableNi", TRUE);
	else
		dbus_setter_int (&priv->g_conn, "SetSuplEnableNi", FALSE);
}

extern void on_interval_spin_button_change_value (GtkSpinButton * spinbutton,
												  gpointer user_data)
{
	dbus_setter_int (&priv->g_conn, "SetFixInterval",
					 gtk_spin_button_get_value_as_int (spinbutton));
	dbus_getter_int (&priv->g_conn, "GetFixInterval");
}

extern void on_single_shot_checkbutton_toggled (GtkToggleButton * togglebutton,
												gpointer user_data)
{
	gint single_shot;
	
	if (gtk_toggle_button_get_active (togglebutton))
		single_shot = 1;
	else
		single_shot = 0;
	
	dbus_setter_int (&priv->g_conn, "SetSingleShot", single_shot);
}

extern void on_supl_server_auto_checkbutton_toggled (GtkToggleButton *togglebutton,
													 gpointer user_data)
{
	gint auto_configure;
	
	gtk_widget_set_sensitive (priv->supl_server_entry,
							  (!gtk_toggle_button_get_active (togglebutton)
							   &&
							   gtk_combo_box_get_active (GTK_COMBO_BOX
														 (priv->mode_combo_box)) > 0));
	
	if (gtk_toggle_button_get_active (togglebutton))
		auto_configure = 1;
	else
		auto_configure = 0;
	
	dbus_setter_int (&priv->g_conn, "SetSuplAutoConfigureAddress", auto_configure);
}

extern void on_disclaimer_dialog_response (GtkDialog * dialog, gint response_id,
										   gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET (dialog));
}

void on_mode_combo_box_changed (GtkComboBox * combo_box, gpointer user_data)
{
	gint assisted;
	GError *error = NULL;

	assisted = gtk_combo_box_get_active (combo_box);

	if (assisted == 1)
		assisted = GPS_MODE_SUPL;
	else
		assisted = GPS_MODE_S_ALONE;

	switch (assisted) {
	case GPS_MODE_OFF:
		break;

	case GPS_MODE_S_ALONE:
		set_supl_settings_sensitivity (FALSE);
		break;

	case GPS_MODE_SUPL:
		set_supl_settings_sensitivity (TRUE);
		break;

	default:
		/* Should never happen */
		g_error ("Illegal GPS Mode");
		g_free (error);
		return;
		break;
	}

	dbus_setter_int (&priv->g_conn, "SetPrefGpsMode", assisted);
	dbus_getter_int (&priv->g_conn, "GetPrefGpsMode");
}

extern void on_systray_menu_showhide (void)
{
	gboolean visible;

	g_object_get (priv->window, "visible", &visible, NULL);
	if (visible) {
		gtk_widget_hide (priv->window);
	} else {
		gtk_widget_show (priv->window);
	}
}

extern void on_systray_menu_quit (void)
{
	gtk_main_quit ();
}

void on_window_destroy (GtkObject * object, gpointer user_data)
{
	gtk_main_quit ();
}

static void on_supl_uncert_dialog_response (GtkDialog * dialog,
											gint response_id,
											gpointer user_data)
{
	int allow = 0;

	if (response_id == GTK_RESPONSE_DELETE_EVENT)
		priv->accept_unknown_certificate = 0;
	else if (response_id == GTK_RESPONSE_YES) {
		allow = 1;
		priv->accept_unknown_certificate = 1;
	} else if (response_id == GTK_RESPONSE_NO)
		priv->accept_unknown_certificate = 0;
	else
		g_debug ("The response is unknown.\n");

	send_supl_uncert_reply (priv->nmea_fd, priv->current_uncert_msg_id, allow);

	gtk_widget_destroy (GTK_WIDGET (dialog));
	priv->supl_uncert_dialog_showing = 0;
}

static void on_supl_ni_dialog_response (GtkDialog * dialog, gint response_id,
										gpointer user_data)
{
	int response_required = 0;
	int allow = 0;

	if (response_id == GTK_RESPONSE_DELETE_EVENT) {
		allow = 0;
		response_required = 1;
	} else if (response_id == GTK_RESPONSE_YES) {
		allow = 1;
		response_required = 1;
	} else if (response_id == GTK_RESPONSE_NO) {
		allow = 0;
		response_required = 1;
	} else if (response_id == GTK_RESPONSE_OK)
		response_required = 0;
	else
		g_debug ("The response is unknown.\n");

	if (priv->current_ni_request_msg_id == -1)
		response_required = 0;

	if (response_required)
		send_supl_ni_reply (priv->nmea_fd, priv->current_ni_request_msg_id, allow);

	gtk_widget_destroy (GTK_WIDGET (dialog));
	priv->supl_ni_dialog_showing = 0;
}

static void on_status_icon_activated ()
{
	if (GTK_WIDGET_VISIBLE (priv->window))
		gtk_widget_hide (priv->window);
	else
		gtk_widget_show (priv->window);
}

static void on_status_icon_popup_menu (GtkStatusIcon * icon, guint button,
									   guint activate_time, gpointer user_data)
{
	gtk_menu_popup (GTK_MENU (priv->systray_menu), NULL, NULL,
					set_systray_menu_position, user_data, button,
					activate_time);
}

static void show_supl_uncert_dialog (char *message, int message_id)
{
	if (priv->supl_ni_dialog_showing) {
		gtk_widget_destroy (GTK_WIDGET (priv->supl_ni_dialog));
		priv->supl_ni_dialog_showing = 0;
	}

	priv->supl_uncert_dialog =
		gtk_message_dialog_new (GTK_WINDOW (priv->window), GTK_DIALOG_MODAL,
								GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
								"%s", message);
	gtk_window_set_title (GTK_WINDOW (priv->supl_uncert_dialog),
						  _("Unknown certificate"));
	g_signal_connect_swapped (priv->supl_uncert_dialog, "response",
							  G_CALLBACK (on_supl_uncert_dialog_response),
							  priv->supl_uncert_dialog);
	gtk_widget_show (priv->supl_uncert_dialog);
	priv->supl_uncert_dialog_showing = 1;
}

static void show_supl_ni_dialog (char *message, int mode, int message_id)
{
	char full_message[MAX_MSG_SIZE];
	GtkMessageType message_type;
	GtkButtonsType buttons_type;
	GtkDialogFlags flags;

	if (priv->supl_ni_dialog_showing) {
		gtk_widget_destroy (GTK_WIDGET (priv->supl_ni_dialog));
		priv->supl_ni_dialog_showing = 0;
	}

	strcpy (full_message, message);

	if (mode == SUPL_NI_ALLOW_ON_TIMEOUT) {
		strcat (full_message,
				_("\n\nIf you don't answer, it will be allowed."));
		message_type = GTK_MESSAGE_QUESTION;
		buttons_type = GTK_BUTTONS_YES_NO;
		flags = GTK_DIALOG_MODAL;
		priv->supl_ni_dialog_showing = 1;
	} else if (mode == SUPL_NI_DENY_ON_TIMEOUT) {
		strcat (full_message,
				_("\n\nIf you don't answer, it will not be allowed."));
		message_type = GTK_MESSAGE_QUESTION;
		buttons_type = GTK_BUTTONS_YES_NO;
		flags = GTK_DIALOG_MODAL;
		priv->supl_ni_dialog_showing = 1;
	} else if (mode == SUPL_NI_NOTIFICATION) {
		message_type = GTK_MESSAGE_INFO;
		buttons_type = GTK_BUTTONS_OK;
		flags = GTK_DIALOG_DESTROY_WITH_PARENT;
		priv->supl_ni_dialog_showing = 0;
	} else {
		g_debug ("Unknown notification type: %d.\n", mode);
		return;
	}
	priv->supl_ni_dialog =
		gtk_message_dialog_new (GTK_WINDOW (priv->window), flags,
								message_type, buttons_type, "%s", full_message);
	gtk_window_set_title (GTK_WINDOW (priv->supl_ni_dialog),
						  _("Network position request"));
	g_signal_connect_swapped (priv->supl_ni_dialog, "response",
							  G_CALLBACK (on_supl_ni_dialog_response),
							  priv->supl_ni_dialog);
	gtk_widget_show (priv->supl_ni_dialog);

}

static void setup_settings_frame ()
{
	gtk_box_pack_start (GTK_BOX (priv->settings_vbox),
						priv->supl_settings_vbox,
						TRUE,
						TRUE,
						0);
	gtk_frame_set_label (GTK_FRAME (priv->settings_frame), _("SUPL settings"));
}

static void init_gui (DBUS_Connection * conn)
{
	int mode, fix_int, allow_ni, auto_configure, single_shot;
	int gps_cust, gps_supl_cust;
	gchar *supl_apn, *supl_server, *supl_user, *supl_password;
	int assisted_tab;

	mode = dbus_getter_int (conn, "GetPrefGpsMode");
	gps_cust = dbus_getter_int (conn, "GetGpsCustomization");
	gps_supl_cust = dbus_getter_int (conn, "GetGpsSuplCustomization");

	if (!gps_cust) {
		g_debug ("Your module is not customized for GPS. Exiting.\n");
		gtk_main_quit ();
		exit (1);
	}

	/* check if lto or supl is enabled in the customization. If not, remove their settings */
	if (gps_supl_cust) {
		if (mode > 1) {
			set_supl_settings_sensitivity (TRUE);
			gtk_combo_box_set_active (GTK_COMBO_BOX (priv->mode_combo_box), 1);
		} else {
			set_supl_settings_sensitivity (FALSE);
			gtk_combo_box_set_active (GTK_COMBO_BOX (priv->mode_combo_box), 0);
		}
	} else {
		gtk_combo_box_set_active (GTK_COMBO_BOX (priv->mode_combo_box), 0);
		gtk_combo_box_remove_text (GTK_COMBO_BOX (priv->mode_combo_box), 1);
		/* remove notebook tab */
		assisted_tab =
			gtk_notebook_page_num (GTK_NOTEBOOK (priv->notebook), priv->assisted_alignment);
		gtk_notebook_remove_page (GTK_NOTEBOOK (priv->notebook), assisted_tab);
	}

	fix_int = dbus_getter_int (conn, "GetFixInterval");

	gtk_spin_button_set_value (GTK_SPIN_BUTTON (priv->interval_spin_button),
							   (gdouble) fix_int);

	/* only get supl settings if supl is enabled in customization */
	if (gps_supl_cust) {
		supl_apn = dbus_getter_str (conn, "GetSuplApn");
		gtk_entry_set_text (GTK_ENTRY (priv->supl_apn_entry), supl_apn);

		auto_configure = dbus_getter_int (conn, "GetSuplAutoConfigureAddress");
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
									  (priv->supl_server_auto_checkbutton),
									  auto_configure);

		supl_server = dbus_getter_str (conn, "GetSuplAddress");
		gtk_entry_set_text (GTK_ENTRY (priv->supl_server_entry), supl_server);

		supl_user = dbus_getter_str (conn, "GetSuplUser");
		gtk_entry_set_text (GTK_ENTRY (priv->supl_user_entry), supl_user);

		supl_password = dbus_getter_str (conn, "GetSuplPassword");
		gtk_entry_set_text (GTK_ENTRY (priv->supl_password_entry), supl_password);

		allow_ni = dbus_getter_int (conn, "GetSuplEnableNi");
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
									  (priv->supl_allow_ni_checkbutton), allow_ni);

		single_shot = dbus_getter_int (conn, "GetSingleShot");
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
									  (priv->single_shot_checkbutton), single_shot);
	}

	setup_settings_frame ();
	gtk_window_resize (GTK_WINDOW (priv->window), 1, 1);
}

void init (int argc, char *argv[])
{
	GtkBuilder *builder;

	priv->maxfd = -1;
	priv->connection_failed = 0;
	priv->current_uncert_msg_id = -1;
	priv->current_uncert_msg_id = -1;
	priv->current_ni_request_msg_id = -1;
	priv->supl_ni_dialog_showing = 0;
	priv->supl_uncert_dialog_showing = 0;
	priv->accept_unknown_certificate = 0;
	priv->old_gps_mode = 0;
	priv->supl_failed = 0;
	priv->no_fix_dialog_displayed = FALSE;
	
	gtk_init (&argc, &argv);

	builder = gtk_builder_new ();

	gtk_builder_add_from_file (builder,
							   "/usr/share/mbm-gps-control/gps_control.xml", NULL);

	priv->window = GTK_WIDGET (gtk_builder_get_object (builder, "window"));
	priv->notebook = GTK_WIDGET (gtk_builder_get_object (builder, "notebook"));
	priv->settings_vbox =
		GTK_WIDGET (gtk_builder_get_object (builder, "settings_vbox"));
	priv->assisted_alignment =
		GTK_WIDGET (gtk_builder_get_object (builder, "assisted_alignment"));
	priv->start_button =
		GTK_WIDGET (gtk_builder_get_object (builder, "start_button"));
	priv->stop_button = GTK_WIDGET (gtk_builder_get_object (builder, "stop_button"));
	priv->single_shot_checkbutton =
		GTK_WIDGET (gtk_builder_get_object
					(builder, "single_shot_checkbutton"));
	gtk_widget_set_sensitive (priv->stop_button, FALSE);
	priv->mode_combo_box =
		GTK_WIDGET (gtk_builder_get_object (builder, "mode_combo_box"));
	gtk_combo_box_set_active (GTK_COMBO_BOX (priv->mode_combo_box), 0);
	priv->interval_spin_button =
		GTK_WIDGET (gtk_builder_get_object (builder, "interval_spin_button"));
	priv->statusbar = GTK_WIDGET (gtk_builder_get_object (builder, "statusbar"));
	priv->statusbar_context =
		gtk_statusbar_get_context_id (GTK_STATUSBAR (priv->statusbar), "Statusbar");
	gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
						"GPS is off");
	priv->status_image = GTK_WIDGET (gtk_builder_get_object (builder, "status_image"));
	priv->systray_menu =
		GTK_WIDGET (gtk_builder_get_object (builder, "systray_menu"));

	priv->status_icon =
		gtk_status_icon_new_from_file ("/usr/share/mbm-gps-control/GPSDisconnected.png");
	gtk_status_icon_set_visible (priv->status_icon, TRUE);
	g_signal_connect (G_OBJECT (priv->status_icon), "activate",
					  G_CALLBACK (on_status_icon_activated), NULL);
	g_signal_connect (G_OBJECT (priv->status_icon), "popup-menu",
					  G_CALLBACK (on_status_icon_popup_menu), NULL);

	/* supl settings widgets */
	priv->enable_supl_radio_button =
		GTK_WIDGET (gtk_builder_get_object
					(builder, "enable_supl_radio_button"));
	priv->settings_frame =
		GTK_WIDGET (gtk_builder_get_object (builder, "settings_frame"));
	priv->supl_settings_vbox =
		GTK_WIDGET (gtk_builder_get_object (builder, "supl_settings_vbox"));
	g_object_ref (priv->supl_settings_vbox);
	priv->supl_apn_entry =
		GTK_WIDGET (gtk_builder_get_object (builder, "supl_apn_entry"));
	priv->supl_server_entry =
		GTK_WIDGET (gtk_builder_get_object (builder, "supl_server_entry"));
	priv->supl_user_entry =
		GTK_WIDGET (gtk_builder_get_object (builder, "supl_user_entry"));
	priv->supl_password_entry =
		GTK_WIDGET (gtk_builder_get_object (builder, "supl_password_entry"));
	priv->supl_allow_ni_checkbutton =
		GTK_WIDGET (gtk_builder_get_object
					(builder, "supl_allow_ni_checkbutton"));
	priv->supl_server_auto_checkbutton =
		GTK_WIDGET (gtk_builder_get_object
					(builder, "supl_server_auto_checkbutton"));

	/* log window */
	log_window = GTK_WIDGET (gtk_builder_get_object (builder, "log_window"));
	nmea_data_text_view =
		GTK_WIDGET (gtk_builder_get_object (builder, "nmea_data_text_view"));

	priv->latitude_label =
		GTK_WIDGET (gtk_builder_get_object (builder, "latitude_label"));
	priv->longitude_label =
		GTK_WIDGET (gtk_builder_get_object (builder, "longitude_label"));
	priv->altitude_label =
		GTK_WIDGET (gtk_builder_get_object (builder, "altitude_label"));
	priv->fix_quality_label =
		GTK_WIDGET (gtk_builder_get_object (builder, "fix_quality_label"));

	supl_cert_init_window (builder, &priv->g_conn);

	gtk_builder_connect_signals (builder, NULL);
	g_object_unref (G_OBJECT (builder));

	gtk_widget_show (priv->window);
}

void start_gps ()
{
	gint readback = FALSE;

	gtk_widget_set_sensitive (priv->stop_button, TRUE);
	gtk_widget_set_sensitive (priv->start_button, FALSE);

	dbus_setter_int (&priv->g_conn, "SetGpsEnabled", TRUE);
	gtk_statusbar_pop (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context);

	priv->no_fix_dialog_displayed = FALSE;

        if (gtk_combo_box_get_active (GTK_COMBO_BOX (priv->mode_combo_box)) > 0)
            priv->supl_selected = TRUE;
        else
            priv->supl_selected = FALSE;

        readback = dbus_getter_int (&priv->g_conn, "GetGpsEnabled");

	if (!readback) {
		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
							_("GPS is off"));
		return;
	}

	gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->statusbar_context,
						_("GPS is on"));
}

void stop_gps ()
{
	gtk_widget_set_sensitive (priv->start_button, TRUE);
	gtk_widget_set_sensitive (priv->stop_button, FALSE);
	dbus_setter_int (&priv->g_conn, "SetGpsEnabled", FALSE);
	dbus_getter_int (&priv->g_conn, "GetGpsEnabled");
	priv->valid_fix = 0;
}

static void init_internationalization ()
{
	setlocale (LC_ALL, "");
	bindtextdomain (GETTEXT_PACKAGE, MBMLOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
}

int main (int argc, char *argv[])
{
	init_internationalization ();

	priv = malloc (sizeof (struct MBMGuiPrivate));
	
	dbus_connect (&priv->g_conn);
	init (argc, argv);
	init_gui (&priv->g_conn);
	open_nmea_fd ();
	g_timeout_add (1000, update_gps_status, &priv->g_conn);
	g_timeout_add (10, read_nmea, &priv->g_conn);
	g_timeout_add (1000, check_time_without_fix, &priv->g_conn);

	gtk_main ();

	dbus_disconnect (&priv->g_conn);

	return 0;
}
